import typer

app = typer.Typer()

SAMPLE_CONFIG = """#!/usr/bin/env python
from __future__ import annotations

import datetime
import re
from decimal import Decimal

from ledger_importer import Config, runner, Posting, Amount

# Custom ledger importer configuration
class LedgerImporterConfig(Config):
    # Define the number of lines that needs to be skipped at the beginning of the file.
    # This is usefull if the csv has a line with the column names for example.
    skip_lines: int = 1

    # Define the csv delimiter
    csv_delimiter: str = ","

    # The argument `fields` given in all parse_* methods contains a whole csv row in a tuple
    # Each element of the tuple is a string representation of the column

    def parse_date(self, fields: tuple) -> datetime.datetime:
        return datetime.datetime.strptime(fields[0], "%m-%d-%Y")

    def parse_payee(self, fields: tuple) -> str:
        return fields[2]

    def parse_postings(self, fields: tuple) -> list[Posting]:
        amount = Amount(quantity=Decimal(re.sub("[€ ]", "", fields[3]).replace(",", ".")), commodity="€")
        if amount > 0:
            account = "Income"
        else:
            account = "Expenses"

        posting = Posting(account="Assets:Checking", amount=amount)

        # A Transaction should be at least 2 postings that have reversed amounts
        return [posting, Posting(account=account, amount=amount.reverse())]


# The next lines are required to run ledger_importer
# when the config file is executed.
if __name__ == "__main__":
    runner(LedgerImporterConfig())
"""


@app.command("init")
def init():
    """
    Bootstrap a config file that can later be customized.
    """
    print(SAMPLE_CONFIG)


if __name__ == "__main__":
    app()



from __future__ import annotations

import csv
import pathlib
import re
import readline
from typing import Optional

import typer

from ledger_importer.config import Config
from ledger_importer.transaction import Amount
from ledger_importer.transaction import Posting
from ledger_importer.transaction import Transaction
from ledger_importer.transactions_handler import TransactionsHandler

__all__ = ("Config", "runner", "Amount", "Posting", "Transaction")


ACCOUNT_LINE = r"^account ([A-z_ :]*[A-z])( +;?.*)?"


def parse_accounts(data: str) -> list[str]:
    accounts: list[str] = []

    for line in data.split("\n"):
        m = re.match(ACCOUNT_LINE, line)
        if not m:
            continue
        accounts.append(m.groups()[0])

    return accounts


class AccountCompleter:
    def __init__(self, accounts: list[str]) -> None:
        self.accounts = accounts

    def complete(self, text: str, state: int) -> str | None:
        # on first trigger, build possible matches
        if state == 0:
            if not text:
                self.matches = self.accounts[:]
            else:
                self.matches = [s for s in self.accounts if s and s.startswith(text)]

        try:
            return self.matches[state]
        except IndexError:
            pass

        return None


def main_wrapper(config: Config):
    def main(
        csv_path: pathlib.Path = typer.Argument(..., help="Path to the bank statement to import."),
        journal_path: Optional[pathlib.Path] = typer.Option(
            None, help="Path a ledger journal to write & learn accounts from."
        ),
        quiet: bool = typer.Option(False, help="Don't ask questions and guess all the accounts automatically."),
    ):
        """
        Import a bank statement.
        """
        # Get accounts list
        accounts: list[str] = []
        if journal_path:
            with journal_path.open() as journal_file:
                accounts = parse_accounts(journal_file.read())

        # Setup account completion
        completer = AccountCompleter(accounts)
        readline.set_completer(completer.complete)
        readline.parse_and_bind("tab: complete")
        readline.set_completer_delims(" \t\n;")

        # Parse transactions, merge them and confirm them
        handler = TransactionsHandler(config)

        with csv_path.open() as csv_file:
            transactions = handler.merge_transactions(
                handler.parse_transactions(
                    csv.reader(csv_file, delimiter=config.csv_delimiter),
                ),
            )
            if not quiet:
                transactions = handler.confirm_transactions(transactions)

        for transaction in transactions:
            print(transaction.to_ledger())

    return main


def runner(config: Config):
    typer.run(main_wrapper(config))



from __future__ import annotations

import datetime
from dataclasses import dataclass
from decimal import Decimal


@dataclass(frozen=True)
class Amount:
    quantity: Decimal
    commodity: str

    def __str__(self) -> str:
        return f"{self.quantity} {self.commodity}"

    def reverse(self) -> Amount:
        return Amount(quantity=-self.quantity, commodity=self.commodity)

    def __gt__(self, val) -> bool:
        return self.quantity > val


@dataclass
class Posting:
    account: str
    amount: Amount


@dataclass
class Transaction:
    """
    Representation of a Transaction.
    """

    date: datetime.datetime
    payee: str
    postings: list[Posting]

    def to_ledger(self):
        return (
            f"{self.date.strftime('%Y/%m/%d')}    {self.payee}\n"
            + "\n".join([f"    {posting.account}    {posting.amount}" for posting in self.postings])
            + "\n"
        )



from __future__ import annotations

import sys

import _csv

from ledger_importer.config import Config
from ledger_importer.transaction import Amount
from ledger_importer.transaction import Posting
from ledger_importer.transaction import Transaction


class TransactionsHandler:
    config: Config

    def __init__(self, config: Config) -> None:
        self.config = config

    def row_to_transaction(self, row: tuple) -> Transaction:
        return Transaction(
            date=self.config.parse_date(row),
            payee=self.config.parse_payee(row),
            postings=self.config.parse_postings(row),
        )

    def merge_transactions(self, transactions: list[Transaction]) -> list[Transaction]:
        """
        Take a list of transactions already sorted by date (oldest -> newest).
        Update the target_accounts of the transactions when they seem to match.

        The matching is tried only on positive transactions, in this case
        it will look for older matching transactions.
        """
        merged_transactions: list[Transaction] = []

        for transaction in transactions:
            if transaction.postings[0].amount > 0:
                for matching_transaction in transactions:
                    if any(matching_transaction is t for t in merged_transactions):
                        # Transaction can be merged only once
                        continue

                    if matching_transaction.date > transaction.date:
                        # Nothing was found up to the current transaction
                        break

                    if self.config.transactions_match(transaction, matching_transaction):
                        transaction.postings[1].account = matching_transaction.postings[0].account
                        merged_transactions.append(matching_transaction)
                        break

        return [transaction for transaction in transactions if transaction not in merged_transactions]

    def parse_transactions(self, csv_reader: _csv._reader) -> list[Transaction]:
        """
        Parse transactions from the csv reader and sort them chronologically.
        """
        transactions: list[Transaction] = []

        for _ in range(self.config.skip_lines):
            next(csv_reader)

        for row in csv_reader:
            transaction = self.row_to_transaction(tuple(row))
            transactions.append(transaction)

        return sorted(transactions, key=lambda transaction: transaction.date)

    def confirm_transactions(self, transactions: list[Transaction]) -> list[Transaction]:
        """
        Manually confirm the transactions using the cli.

        Returns only confirmed transactions.
        """
        confirmed_transactions: list[Transaction] = []

        for transaction in transactions:
            print(
                f"""
| {"Account".center(max(len(transaction.postings[0].account), len("Account")))} | {"Date".center(10)} | {"Amount".center(max(len(str(transaction.postings[0].amount)), len("Amount")))} | {"Payee".center(max(len(transaction.payee), len("Payee")))} |
| {transaction.postings[0].account} | {transaction.date.strftime("%Y/%m/%d")} | {transaction.postings[0].amount} | {transaction.payee} |
    """,
                file=sys.stderr,
            )

            if transaction.postings[0].amount.quantity > 0:
                print(
                    f"""Which account provided this income? ([{transaction.postings[1].account}]/[q]uit/[s]kip) """,
                    file=sys.stderr,
                )
            else:
                print(
                    f"""To which account did this money go? ([{transaction.postings[1].account}]/[q]uit/[s]kip) """,
                    file=sys.stderr,
                )

            answer = input()
            if answer == "q":
                break
            elif answer == "s":
                continue

            if answer:
                transaction = Transaction(
                    date=transaction.date,
                    payee=transaction.payee,
                    postings=[
                        transaction.postings[0],
                        Posting(
                            account=answer,
                            amount=Amount(
                                quantity=transaction.postings[1].amount.quantity,
                                commodity=transaction.postings[1].amount.commodity,
                            ),
                        ),
                    ],
                )
            confirmed_transactions.append(transaction)

        return confirmed_transactions



from __future__ import annotations

import datetime
from abc import ABC
from abc import abstractmethod

from ledger_importer.transaction import Posting
from ledger_importer.transaction import Transaction


class Config(ABC):
    skip_lines: int = 1
    csv_delimiter: str = ","

    @abstractmethod
    def parse_date(self, fields: tuple) -> datetime.datetime:
        pass

    @abstractmethod
    def parse_payee(self, fields: tuple) -> str:
        pass

    @abstractmethod
    def parse_postings(self, fields: tuple) -> list[Posting]:
        pass

    def transactions_match(self, transaction1: Transaction, transaction2: Transaction) -> bool:
        """
        Matching transactions will get merged in a single transaction.

        When they are merged, the target_account of transaction1 will be updated to the account of transaction2. This overrides the parse_target_account return value.

        Default behavior is to merge transactions with opposite amounts and different accounts.
        """
        return (
            transaction1.postings[0].amount == transaction2.postings[0].amount.reverse()
            and transaction1.postings[0].account != transaction2.postings[0].account
        )
